/*==========================================================================*\
 |  This file is part of setiherder (http://www.sackheads.org/setiherder/)  |
 |  which was written by Brian L. Naylor <brian+seti@sackheads.org> and is  |
 |  under the GNU General Public License.  You should have recieved a copy  |
 |  of the GPL with this source code.  If not, we'd like to know about it.  |
\*--------------------------------------------------------------------------*/

#ifndef SH_NET
#define SH_NET

/*==========================================================================*\
 |                                                                          |
 | Extremely simple protocol for setiherder<->setid interaction:            |
 |                                                                          |
 | All network commands and responses are prefixed with SH_NET_PREFIX,      |
 | anything else is 1) setid->setiherder only and 2) assumed to be output   |
 | from the seti@home client.  This prefix allows for quick and easy        |
 | parsing on the gui side (for protocol traffic.)                          |
 |                                                                          |
 | Note that each command is matched simply for the initial token.  Extra   |
 | trailing information is allowed, up to the max buffer length.  This can  |
 | be used for extra information on an error, or in the case of 'detach',   |
 | a mandatory command argument.                                            |
 |                                                                          |
 | When the gui sends a command to the daemon, the daemon will *always*     |
 | respond.  The daemon, however, does not expect this from the gui and     |
 | will simply discard return codes and so forth.                           |
 |                                                                          |
 | The gui should be prepared to receive un-asked-for status codes, such    |
 | as 'died', etc.                                                          |
 |                                                                          |
 | Commands:                                                                |
 |                                                                          |
 |   SH_NET_CMD_START:                                                      |
 |      gui->daemon, request to start seti@home client.                     |
 |      possible responses: ok          : client started                    |
 |                          active      : client is already running         |
 |                          client_start: error starting client             |
 |                                                                          |
 |  SH_NET_CMD_KILL:                                                        |
 |     gui->daemon, request to stop seti@home client.                       |
 |     possible responses:  ok          : client was killed                 |
 |                          nactive     : client isn't running              |
 |                          client_comm : client isn't responding           |
 |     If the operation was successful, a 'died' will come immediately      |
 |     after the response code.                                             |
 |                                                                          |
 |  SH_NET_CMD_DETACH:                                                      |
 |     bidirectional, sets or reports state of 'detach' flag.               |
 |     possible responses:  ok          : flag was set                      |
 |                                                                          |
 |     This command must include a single argument, an integer value.  When |
 |     sent from the gui, this sets the value of the 'detach' flag.  When   |
 |     send from the daemon, it simply reports its current state.  (This is |
 |     done when a gui attaches initially, so that it will know the correct |
 |     state.)                                                              |
 |                                                                          |
 |     Flag values: 0: no-detach - this is the default, and will cause the  |
 |                                 seti@home client to be killed when the   |
 |                                 setiherder gui closes its tcp connection |
 |                                 to the daemon.                           |
 |                  1: detach    - this will cause the daemon to let the    |
 |                                 client continue running after the gui    |
 |                                 closes connection, and report extra      |
 |                                 information when it reattaches.          |
 |                  other: undefined, but will currently count as 1.        |
 |                                                                          |
 |  SH_NET_CMD_PING:                                                        |
 |     gui->daemon, tests communication.                                    |
 |     possible responses: pong          : I'm alive                        |
 |                                                                          |
 |  SH_NET_CMD_IDLE:                                                        |
 |     daemon->gui, reports work done while gui was disconnected            |
 |     possible responses: none                                             |
 |                                                                          |
 |     This token will be followed with a single integer argument           |
 |     representing the work units completed by the client since the gui    |
 |     was last connected.  This value will be immediately reset by the     |
 |     daemon.                                                              |
 |                                                                          |
 |  SH_NET_CMD_AUTH:                                                        |
 |     gui->daemon, issues plaintext password                               |
 |     possible responses: ok      : good passwd                            |
 |                         authreq : bad passwd                             |
 |                                                                          |
 |  SH_NET_CMD_VERSION:                                                     |
 |     gui->daemon, asks for software/protocol versions                     |
 |     possible responses: version : lists versions                         |
 |                                                                          |
 |     This token will be followed by two strings, the first being the      |
 |     setiherder release version, the second being the protocol version.   |
 |     (The software could be released without there being a change in the  |
 |     protocol..  and probably will be.)                                   |
 |                                                                          |
 | Status codes:                                                            |
 |  SH_NET_ERR_OK:           success                                        |
 |  SH_NET_ERR_PONG:         response to ping command                       |
 |  SH_NET_ERR_ACTIVE:       client is already active                       |
 |  SH_NET_ERR_NACTIVE:      client is not active                           |
 |  SH_NET_ERR_CONNECT:      another gui is connected                       |
 |  SH_NET_ERR_CLIENT_START: client wouldn't start                          |
 |  SH_NET_ERR_CLIENT_COM:   client is not responding                       |
 |  SH_NET_ERR_HOST:         you are not allowed to connect                 |
 |  SH_NET_ERR_DIED:         the client has terminated                      |
 |  SH_NET_ERR_SHUTDOWN:     setid is being killed                          |
 |  SH_NET_ERR_AUTHREQ:      you need to authenticate                       |
 |  SH_NET_ERR_VERSION:      lists versions                                 |
 |  SH_NET_ERR_SYNTAX:       I DO NOT UNDERSTAND YOUR CRAZY MOONSPEAK!!     |
 |  SH_NET_ERR_GENERIC:      undefined error                                |
 |                                                                          |
\*==========================================================================*/

#include "shtypes.h"

#define SH_NET_VERSION             "1.1"
#define SH_NET_VERSION_MAJOR       1
#define SH_NET_VERSION_MINOR       1

#define SH_NET_DEFAULT_PORT        7501

#define SH_NET_PREFIX              "SD:"

#define SH_NET_CMD_START           "start"
#define SH_NET_CMD_KILL            "kill"
#define SH_NET_CMD_DETACH          "detach"
#define SH_NET_CMD_PING            "ping"
#define SH_NET_CMD_IDLE            "idle"
#define SH_NET_CMD_AUTH            "auth"
#define SH_NET_CMD_VERSION         "version"

#define SH_NET_ERR_OK              "success"
#define SH_NET_ERR_PONG            "pong"
#define SH_NET_ERR_ACTIVE          "active"
#define SH_NET_ERR_NACTIVE         "nactive"
#define SH_NET_ERR_CONNECT         "connect"
#define SH_NET_ERR_CLIENT_START    "client_start"
#define SH_NET_ERR_CLIENT_COM      "client_com"
#define SH_NET_ERR_HOST            "host"
#define SH_NET_ERR_DIED            "died"
#define SH_NET_ERR_SHUTDOWN        "shutdown"
#define SH_NET_ERR_AUTHREQ         "authreq"
#define SH_NET_ERR_VERSION         "version"
#define SH_NET_ERR_SYNTAX          "confused"
#define SH_NET_ERR_GENERIC         "generic"


int   shnet_init(void);
int   shnet_connect(sh_client_t *);
int   shnet_close(sh_client_t *);
void  shnet_idle(void);
int   shnet_parse(sh_client_t *, char *);
int   shnet_send_cmd(sh_client_t *, char *);
int   shnet_set_detach(sh_client_t *);
int   shnet_send_passwd(sh_client_t *);
short shnet_next_seq(int);
int   shnet_ping(sh_client_t *);

#endif
