/*==========================================================================*\
 |  This file is part of setiherder (http://www.sackheads.org/setiherder/)  |
 |  which was written by Brian L. Naylor <brian+seti@sackheads.org> and is  |
 |  under the GNU General Public License.  You should have recieved a copy  |
 |  of the GPL with this source code.  If not, we'd like to know about it.  |
\*--------------------------------------------------------------------------*/

#include "shqueue.h"
#include "shnet.h"
#include "shwin.h"
#include "shutil.h"
#include "shclients.h"

/* 
 * globals pertinent to this module 
 */
static int sh_q_initialized = 0;
sh_q_item_t *queues[SH_MAX_CLIENTS];

/*
 * local utility functions
 */
int shq_nullcheck(idx)
{
  if(idx < 0 || idx > SH_MAX_CLIENTS)
    return 1;

  if(queues[idx] == NULL)
    return 2;

  return 0;
}

int sh_q_init(void)
{
  memset(queues, 0, sizeof(sh_q_item_t*) * SH_MAX_CLIENTS); 
  sh_q_initialized = 1;
  return SH_SUCCESS;
}

void sh_q_cleanup(void)
{
  int          i;
  sh_q_item_t *runner, *tibet;

  for(i=0; i<SH_MAX_CLIENTS; i++)
  {
    runner = queues[i];
    while(runner != NULL) 
    {
      tibet = runner;
      runner = runner->next;
      free(tibet);
    }
  }
  sh_q_initialized = 0;
}

int sh_q_append(int idx, sh_q_item_t *newbie)
{
  sh_q_item_t *runner;
  int          rc = shq_nullcheck(idx);

  if(rc == 1)
    return SH_ERR_Q_NULL;

  if(rc ==2)
  {
    queues[idx] = newbie;
  }
  else
  {
    runner = queues[idx];
    while(runner->next != NULL)
      runner = runner->next;
    runner->next = newbie;
  }
  
  return SH_SUCCESS;
}

/* prepend an item to the queue for client n */
int sh_q_prepend(int idx, sh_q_item_t *newbie)
{
  int rc = shq_nullcheck(idx);

  if(rc == 1)
    return SH_ERR_Q_NULL;

  if(rc == 0)
    newbie->next = queues[idx];

  queues[idx] = newbie;

  return SH_SUCCESS;
}

/* handle input for client n */
int sh_q_got_input(int idx, int seq, sh_q_item_t **orig)
{
  sh_q_item_t *runner, *prev;
  sh_client_t *sc;
  int rc = shq_nullcheck(idx);

  orig = NULL;

  if(rc != SH_SUCCESS)
  {
    return SH_ERR_Q_NULL;
  }

  /*
   * if we previously marked this client as LAGged, reset it
   * (since this is clearly no longer true..)
   */
  sc = sh_client_get(idx);
  if(sc->state == SH_STATE_LAG)
  {
    sc->state = SH_STATE_WORKING;
    shwin_status_change(idx);
    dbg(DBG_INFO, "Client %d (%s) has recovered.\n",
        idx, sc->hostname);
  }

  /* 
   * it really *should* be the first item unless that is a ping..  but
   * we'll check them all anyway..
   */
  for( runner = queues[idx] , prev = runner   ;
      (runner != NULL) && (runner->seq != seq);
       prev = runner , runner = runner->next );

  if( runner == NULL || runner->seq != seq)
  {
    dbg(DBG_ERROR, "sh_q_got_input() experienced a sequencing error, "
        "seq=%d, idx=%d.\n", seq, idx);
    return SH_ERR_Q_SEQ;
  }
   
  orig = &runner;
  if(prev == runner)
    queues[idx] = runner->next;
  else
    prev->next = runner->next;
  
  /*
   * transmit next
   */
  if(queues[idx] != NULL && !queues[idx]->sent)
  {
    shnet_raw_send(idx, queues[idx]->text);
    queues[idx]->sent     = 1;
    queues[idx]->idle_cnt = 0;
  }
  
  return SH_SUCCESS;
}

/* check pending queue items (idle loop) */
gint sh_q_idle(gpointer data)
{
  int          i;
  sh_q_item_t *runner;
  char        *msg;

  for(i=0; i<SH_MAX_CLIENTS; i++)
  {
    if(queues[i] != NULL)
    {
      /* transmit unsent */  
      if(!queues[i]->sent)
      {
        shnet_raw_send(i, queues[i]->text);
        queues[i]->sent     = 1;
        queues[i]->idle_cnt = 0;
      }
      else
      {
        queues[i]->idle_cnt++;
      }
 
      /* check for stragglers */
      if(queues[i]->idle_cnt >= SH_Q_MAX_WAIT)
      {
        if(strstr(queues[i]->text, ":ping\n"))
        {
          /* it's already a ping, we have nasty lag! */
          dbg(DBG_WARN, "Client %d (%s) is not responding!\n",
              i, sh_client_get(i)->hostname);
          sh_client_get(i)->state = SH_STATE_LAG;
          shwin_status_change(i);
          /* reset this to give a little breathing room */
          queues[i]->idle_cnt = 0;
        }
        else 
        {
          /* insert a ping packet */
          shnet_ping(sh_client_get(i));
        }
      }
    }
  }  
  return TRUE;
}


